/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef avro_ResolvingReader_hh__
#define avro_ResolvingReader_hh__

#include <stdint.h>
#include <boost/noncopyable.hpp>

#include "Config.hh"
#include "ResolverSchema.hh"
#include "Reader.hh"

namespace avro {

class AVRO_DECL ResolvingReader : private boost::noncopyable
{

  public:

    ResolvingReader(const ResolverSchema &schema, const InputBuffer &in) :
        reader_(in),
        schema_(schema)
    {}

    template<typename T>
    void parse(T &object) {
        schema_.parse(reader_, reinterpret_cast<uint8_t *>(&object));
    }

  private:

    Reader reader_;
    ResolverSchema schema_;
};

} // namespace avro

#endif
